#include <avr/io.h>

#include "synth.h"

.global TIMER1_COMPA_vect
TIMER1_COMPA_vect:

	; We'll preserve SREG in R2, so back up R2 first then store SREG.
	push r2
	in r2, _SFR_IO_ADDR(SREG)
	
	; Bulk-preserve common registers.
	push r0
	push r1
	push r2
	push r3
	push r14
	push r15
	push r16
	push r17
	push r18
	push r19
	push r20
	push r21
	push r22
	push r23
	push r24
	push r25
	push XL
	push XH
	push YL
	push YH
	push ZL
	push ZH
	
	; Output the previously-generated value.
	lds ZL, synth_output_a
	out _SFR_IO_ADDR(OCR0A), ZL
#if SYNTH_VOICE_COUNT > 1
	lds ZL, synth_output_b
#endif
	out _SFR_IO_ADDR(OCR0B), ZL
	
	ldi YH, hi8(synth_voices);
	ldi YL, lo8(synth_voices);
	
	; This will store the generated value
	clr r24
	clr r25
	movw r14, r24
	
	ldi r16, SYNTH_VOICE_COUNT
	mov r3, r16

process_voice:
	
	; Y points to the current voice	
	
	;; Modulation Operator
	
	; X = phase of operator 1
	ldd XL, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 1) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase + 0
	ldd XH, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 1) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase + 1
	
	; R1:R0 = phase delta of operator 1
	ldd r0, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 1) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase_delta + 0
	ldd r1, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 1) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase_delta + 1
	
	; Point Z to sample on waveform
	ldi ZH, hi8(waveforms)
	ldd ZL, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 1) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_shape
	add ZH, ZL
	mov ZL, XH
	
	; Phase += Delta
	add XL, r0
	adc XH, r1
	
	; Save phase back to operator 1
	std Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 1) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase + 0, XL
	std Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 1) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase + 1, XH
	
	; Read sample from waveform
	lpm r18, Z
	
	; Scale by modulator envelope
	ldd r19, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 1) + offsetof_synth_operator_envelope + offsetof_synth_envelope_value + 1
	mulsu r18, r19
	
	; Scale by modulator amplitude
	
	movw r22, r0
	
	ldd r20, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 1) + offsetof_synth_operator_amplitude + 0
	ldd r21, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 1) + offsetof_synth_operator_amplitude + 1
	
	; r19:r18:r17:r16 = r23:r22 * r21:r20
	
	clr r2
	
	muls r23, r21 ; (signed)ah * (signed)bh
	movw r18, r0
	mul r22, r20 ; al * bl
	movw r16, r0
	
	mulsu r23, r20 ; (signed)ah * bl
	;sbc r19, r2
	add r17, r0
	adc r18, r1
	;adc r19, r2
	
	mulsu r21, r22 ; (signed)bh * al
	;sbc r19, r2
	add r17, r0
	adc r18, r1
	;adc r19, r2
	
	;; Carrier Operator
	
	; X = phase of operator 0
	ldd XL, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 0) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase + 0
	ldd XH, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 0) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase + 1
	
	; R1:R0 = phase delta of operator 0
	ldd r0, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 0) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase_delta + 0
	ldd r1, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 0) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase_delta + 1
	
	add XL, r0
	adc XH, r1
	
	std Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 0) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase + 0, XL
	std Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 0) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_phase + 1, XH
	
	; Carrier operator needs to have its phase offset by modulation operator
	add XL, r17
	adc XH, r18
	
	; Point Z to sample on waveform
	ldi ZH, hi8(waveforms)
	ldd ZL, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 0) + offsetof_synth_operator_oscillator + offsetof_synth_oscillator_shape
	add ZH, ZL
	mov ZL, XH
	
	; Read sample from waveform
	lpm r18, Z
	
	; Scale by carrier envelope
	ldd r19, Y + offsetof_synth_voice_patch + offsetof_synth_patch_operators + (sizeof_synth_operator * 0) + offsetof_synth_operator_amplitude + 1;
	mulsu r18, r19

	; Add to running counter
	add r24, r0
	adc r25, r1
	
#if SYNTH_VOICE_COUNT > 1
	; Swap with alternate register set
	movw r0, r24
	movw r24, r14
	movw r14, r0
#endif
	
	; Have we done all the voices?
	dec r3
	breq finished_voices
	
	adiw YL, sizeof_synth_voice
	
	jmp process_voice

finished_voices:

	ldi ZH, 128
	add r25, ZH
	sts synth_output_a, r25
	
	add r15, ZH
	sts synth_output_b, r15
	
	lds XL, envelope_advance
	dec XL
	brne not_envelope_time
	
	lds XL, synth_envelope_ticks_pending
	inc XL
	sts synth_envelope_ticks_pending, XL
	
	ldi XL, 8
not_envelope_time:
	sts envelope_advance, XL
	
	; Bulk-restore common registers.
	pop ZH
	pop ZL
	pop YH
	pop YL
	pop XH
	pop XL
	pop r25
	pop r24
	pop r23
	pop r22
	pop r21
	pop r20
	pop r19
	pop r18
	pop r17
	pop r16
	pop r15
	pop r14
	pop r3
	pop r2
	pop r1
	pop r0
	
	; Restore SREG and R2.
	out _SFR_IO_ADDR(SREG), r2
	pop r2
	reti