#ifndef SYNTH_H
#define SYNTH_H


#define SYNTH_SAMPLE_RATE (36000UL)

#define SYNTH_OPERATOR_COUNT (2)
#define SYNTH_VOICE_COUNT (4)
#define SYNTH_CHANNEL_COUNT (16)

#ifdef __ASSEMBLER__

#include "synth_offsets.h"

#else /* !ASSEMBLER */

#include <stdint.h>

typedef enum __attribute__((packed)) {
	SYNTH_ENVELOPE_STATE_IDLE,
	SYNTH_ENVELOPE_STATE_ATTACK,
	SYNTH_ENVELOPE_STATE_DECAY,
	SYNTH_ENVELOPE_STATE_SUSTAIN,
	SYNTH_ENVELOPE_STATE_RELEASE
} synth_envelope_state;

typedef struct __attribute__((packed)) {
	
	synth_envelope_state state;
	
	uint16_t value;
	
	uint16_t attack;
	uint16_t decay;
	uint16_t sustain;
	uint16_t release;
	
} synth_envelope;

typedef enum __attribute__((packed)) {
	SYNTH_WAVEFORM_SHAPE_SINE = 0,
	SYNTH_WAVEFORM_SHAPE_SQUARE = 1,
	SYNTH_WAVEFORM_SHAPE_TRIANGLE = 2,
	SYNTH_WAVEFORM_SHAPE_SAWTOOTH_UP = 3,
	SYNTH_WAVEFORM_SHAPE_SAWTOOTH_DOWN = 4,
} synth_waveform_shape;

typedef struct __attribute__((packed)) {
	
	synth_waveform_shape shape;
	
	uint16_t phase;
	uint16_t phase_delta;
	int16_t phase_offset;
	
} synth_oscillator;

typedef struct __attribute__((packed)) {
	
	synth_envelope envelope;
	synth_oscillator oscillator;
	uint16_t frequency_scale;
	uint16_t amplitude;
	
} synth_operator;

typedef struct __attribute__((packed)) {
	
	synth_operator operators[SYNTH_OPERATOR_COUNT];
	
} synth_patch;

typedef enum __attribute__((packed)) {
	SYNTH_VOICE_STATE_IDLE,
	SYNTH_VOICE_STATE_ACTIVE,
	SYNTH_VOICE_STATE_ENDING,
} synth_voice_state;

typedef struct __attribute__((packed)) {

	synth_voice_state state;
	
	uint8_t channel;
	uint8_t note;
	uint8_t velocity;
	
	uint8_t age;
	
	synth_patch patch;
	
} synth_voice;

void synth_tick_envelope(synth_envelope *envelope);
void synth_tick_oscillator(synth_oscillator *oscillator);
void synth_set_patch(uint8_t channel, synth_patch *patc);
void synth_begin_voice(synth_voice *voice, uint8_t channel, uint8_t note, uint8_t velocity);
void synth_end_voice(synth_voice *voice);
void synth_set_pitch_bend(uint8_t channel, int16_t amount);


extern volatile uint8_t synth_envelope_ticks_pending;

extern synth_voice synth_voices[SYNTH_VOICE_COUNT];
extern synth_patch synth_patches[SYNTH_CHANNEL_COUNT];

#endif

#endif